from __future__ import annotations

from abilian.sbe.app import Application


def register_plugin(app: Application) -> None:
    from .actions import register_actions
    from .views import wiki

    # could also test on private attribute _got_registered_once
    if not any(
        fct for fct in wiki.deferred_functions if fct.__name__ == "register_actions"
    ):
        wiki.record_once(register_actions)
    app.register_blueprint(wiki)



from __future__ import annotations

import subprocess
import sys
from importlib.metadata import distributions
from operator import itemgetter

from flask import current_app, render_template

from abilian.web.admin import AdminPanel


def installed_packages() -> list[dict[str, str]]:
    packages = [
        {"name": d.name.lower(), "version": d.version, "metadata": {}}
        for d in distributions()
    ]
    return sorted(packages, key=itemgetter("name"))


class SysinfoPanel(AdminPanel):
    id = "sysinfo"
    label = "System information"
    icon = "hdd"

    def get(self) -> str:
        uname = subprocess.check_output(["uname", "-a"]).decode().strip()
        python_version = sys.version.strip()

        packages = installed_packages()

        config_values = [(k, repr(v)) for k, v in sorted(current_app.config.items())]

        ctx = {
            "python_version": python_version,
            "packages": packages,
            "uname": uname,
            "config_values": config_values,
        }
        return render_template("admin/sysinfo.html", **ctx)



"""Models for user preferences."""

from __future__ import annotations

from flask_sqlalchemy import BaseQuery
from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import backref, relation

from abilian.core.extensions import db
from abilian.core.models.subjects import User
from abilian.core.sqlalchemy import JSON


class UserPreference(db.Model):
    """An atom of user preference."""

    __tablename__ = "user_preference"
    __table_args__ = (UniqueConstraint("user_id", "key"),)

    #: Unique id for this preference.
    id = Column(Integer, primary_key=True, autoincrement=True)

    #: The user who set this preference.
    user = relation(
        User, backref=backref("preferences", cascade="all, delete, delete-orphan")
    )
    user_id = Column(ForeignKey(User.id))

    #: The key
    key = Column(String, nullable=False)

    #: The value
    value = Column(JSON, nullable=False)

    query: BaseQuery



"""Folders / Documents module."""

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from abilian.sbe.app import Application


def register_plugin(app: Application) -> None:
    from . import lock, signals
    from .cli import antivirus
    from .models import setup_listener
    from .views import community_blueprint

    app.register_blueprint(community_blueprint)
    setup_listener()

    # set default lock lifetime
    app.config.setdefault("SBE_LOCK_LIFETIME", lock.DEFAULT_LIFETIME)

    app.cli.add_command(antivirus)



# Dead code, not used in SBE (was used in CRM).
# TODO: remove definitively (or move to CRM project).

# """Front-end for a CRM app.
#
# This should eventually allow implementing very custom CRM-style
# application.
# """
#
# from __future__ import annotations
#
# import contextlib
# import copy
# import re
# from collections import OrderedDict
# from collections.abc import Callable, Collection
# from typing import Any
#
# import sqlalchemy as sa
# from flask import (
#     Request,
#     current_app,
#     g,
#     redirect,
#     render_template,
#     request,
#     session,
#     url_for,
# )
# from flask.blueprints import Blueprint, BlueprintSetupState
# from flask_login import current_user
# from sqlalchemy import Date, DateTime, func, orm
# from sqlalchemy.sql.expression import asc, desc, nullsfirst, nullslast
# from werkzeug.exceptions import BadRequest
# from wtforms import Form
#
# from abilian.app import Application
# from abilian.core.entities import Entity, EntityQuery
# from abilian.core.extensions import db
# from abilian.i18n import _l
# from abilian.services import audit_service, get_security_service, get_service
# from abilian.services.security import READ
# from abilian.services.vocabularies.models import BaseVocabulary
#
# from . import search
# from .action import (
#     Action,
#     ActionDropDown,
#     ActionGroup,
#     ActionGroupItem,
#     Endpoint,
#     FAIcon,
#     actions,
# )
# from .forms.widgets import AjaxMainTableView, Panel, RelatedTableView, Row, SingleView
# from .nav import BreadcrumbItem
# from .views import (
#     JSONView,
#     JSONWhooshSearch,
#     ObjectCreate,
#     ObjectDelete,
#     ObjectEdit,
#     ObjectView,
#     default_view,
# )
#
#
# class ModuleAction(Action):
#     """Base action class for :class:`Module` actions.
#
#     Basic condition is simple: :attr:`category` must match the string
#     `'module:{module.endpoint}'`
#     """
#
#     def __init__(self, module: Module, group: str, name: str, *args, **kwargs):
#         self.group = group
#         super().__init__(module.action_category, name, *args, **kwargs)
#
#     def pre_condition(self, context: dict[str, Module]) -> bool:
#         module = actions.context.get("module")
#         if not module:
#             return False
#
#         return self.category == module.action_category
#
#
# class ModuleActionGroup(ModuleAction, ActionGroup):
#     template_string = ActionGroup.template_string
#
#
# class ModuleActionDropDown(ModuleAction, ActionDropDown):
#     template_string = ActionDropDown.template_string
#
#
# class ModuleActionGroupItem(ModuleAction, ActionGroupItem):
#     pass
#
#
#
#
# def expose(url: str = "/", methods: tuple[str] = ("GET",)) -> Callable:
#     """Use this decorator to expose views in your view classes.
#
#     `url`   Relative URL for the view
#
#     `methods` Allowed HTTP methods. By default, only GET is allowed.
#     """
#
#     def wrap(f):
#         if not hasattr(f, "_urls"):
#             f._urls = []
#         f._urls.append((url, methods))
#         return f
#
#     return wrap
#
#
# def make_label(s: str) -> str:
#     return " ".join(w.capitalize() for w in s.split("_"))
#
#
# def make_single_view(form: Form, **options) -> SingleView:
#     panels = []
#     for gr in form._groups.items():
#         panel = Panel(gr[0], *(Row(x) for x in gr[1]))
#         panels.append(panel)
#     return SingleView(form, *panels, **options)
#
#
# class ModuleView:
#     """Mixin for module base views.
#
#     Provide :attr:`module`.
#     """
#
#     #: :class:`Module` instance
#     module: Module
#
#     def __init__(self, module: Module, *args, **kwargs):
#         self.module = module
#         super().__init__(*args, **kwargs)
#
#
# class BaseEntityView(ModuleView):
#     pk = "entity_id"
#
#     form: Form
#
#     def init_object(self, args, kwargs):
#         args, kwargs = super().init_object(args, kwargs)
#
#         # security check
#         if self.obj and not self.check_access():
#             # FIXME: will lead base views to return 404. Ok when permission is
#             # READ, but when it's WRITE would 403 be more appropriate?
#             self.obj = None
#
#         return args, kwargs
#
#     def breadcrumb(self):
#         return BreadcrumbItem(
#             label=self.obj.name or self.obj.id,
#             url=Endpoint(".entity_view", entity_id=self.obj.id),
#         )
#
#     def prepare_args(self, args, kwargs):
#         args, kwargs = super().prepare_args(args, kwargs)
#         actions.context["module"] = self.module
#         add_to_recent_items(self.obj)
#         return args, kwargs
#
#     def redirect_to_index(self):
#         return redirect(self.module.url)
#
#     @property
#     def single_view(self) -> SingleView:
#         return make_single_view(
#             self.form,
#             view_template=self.module.view_template,
#             view=self,
#             module=self.module,
#             **self.module.view_options,
#         )
#
#     def check_access(self):
#         return self._check_view_permission(self)
#
#     def _check_view_permission(self, view):
#         """
#         :param view: a :class:`ObjectView` class or instance
#         """
#         security = get_security_service()
#         return security.has_permission(current_user, view.permission, self.obj)
#
#     @property
#     def can_edit(self):
#         return self._check_view_permission(self.module.edit_cls)
#
#     @property
#     def can_delete(self):
#         return self._check_view_permission(self.module.delete_cls)
#
#     @property
#     def can_create(self):
#         create_cls = self.module.create_cls
#         permission = create_cls.permission
#         cls_permissions = dict(self.Model.__default_permissions__)
#
#         if self.permission in cls_permissions:
#             security = get_service("security")
#             return security.has_permission(
#                 current_user,
#                 create_cls.permission,
#                 obj=self.obj,
#                 roles=cls_permissions[permission],
#             )
#         return False
#
#
# EDIT_ACTION = Action(
#     "module",
#     "object:view",
#     title=_l("Edit"),
#     button="default",
#     condition=lambda ctx: ctx["view"].can_edit,
#     icon=FAIcon("edit"),
#     url=lambda ctx: url_for(".entity_edit", **{ctx["view"].pk: ctx["view"].obj.id}),
# )
#
# DELETE_ACTION = Action(
#     "module",
#     "object:view",
#     title=_l("Delete"),
#     button="danger",
#     condition=lambda ctx: ctx["view"].can_delete,
#     icon=FAIcon("trash fa-inverse"),
#     url=lambda ctx: url_for(".entity_delete", **{ctx["view"].pk: ctx["view"].obj.id}),
# )
# DELETE_ACTION.template = "widgets/frontend_action_delete_confim.html"
#
#
# class EntityView(BaseEntityView, ObjectView):
#     mode = "view"
#     template = "default/single_view.html"
#
#     @property
#     def object_actions(self):
#         ctx = actions.context
#         return [a for a in (EDIT_ACTION, DELETE_ACTION) if a.available(ctx)]
#
#     @property
#     def template_kwargs(self):
#         module = self.module
#         related_views = [v.render(self.obj) for v in module.related_views]
#         rendered_entity = self.single_view.render(self.obj, related_views=related_views)
#         audit_entries = audit_service.entries_for(self.obj)
#
#         return {
#             "rendered_entity": rendered_entity,
#             "related_views": related_views,
#             "audit_entries": audit_entries,
#             "show_new_comment_form": True,
#             "show_new_attachment_form": True,
#             "module": self.module,
#         }
#
#
# class EntityEdit(BaseEntityView, ObjectEdit):
#     template = "default/single_view.html"
#     mode = "edit"
#
#     @property
#     def template_kwargs(self):
#         rendered_entity = self.single_view.render_form()
#         return {
#             "rendered_entity": rendered_entity,
#             "show_new_comment_form": False,
#             "show_new_attachment_form": False,
#             "module": self.module,
#         }
#
#
# class EntityCreate(BaseEntityView, ObjectCreate):
#     template = "default/single_view.html"
#     mode = "create"
#
#     prepare_args = ObjectCreate.prepare_args
#     breadcrumb = ObjectCreate.breadcrumb
#
#     def check_access(self):
#         return self.can_create
#
#     @property
#     def template_kwargs(self):
#         rendered_entity = self.single_view.render_form()
#         return {
#             "rendered_entity": rendered_entity,
#             "for_new": True,
#             "module": self.module,
#         }
#
#
# class EntityDelete(BaseEntityView, ObjectDelete):
#     pass
#
#
# class ListJson(ModuleView, JSONView):
#     """JSON endpoint, for AJAX-backed table views."""
#
#     def data(self, *args, **kwargs) -> dict:
#         echo = int(kwargs.get("sEcho", 0))
#         length = int(kwargs.get("iDisplayLength", 10))
#         start = int(kwargs.get("iDisplayStart", 0))
#         end = start + length
#
#         total_count = self.module.listing_query.count()
#         query = self.module.list_query(request)
#         count = query.count()
#         query = self.module.ordered_query(request, query)
#
#         entities = query.slice(start, end).all()
#         table_view = AjaxMainTableView(
#             columns=self.module.list_view_columns,
#             name=self.module.managed_class.__name__.lower(),
#             ajax_source=url_for(".list_json"),
#         )
#
#         data = [table_view.render_line(e) for e in entities]
#         return {
#             "sEcho": echo,
#             "iTotalRecords": total_count,
#             "iTotalDisplayRecords": count,
#             "aaData": data,
#         }
#
#
# class ModuleMeta(type):
#     """Module metaclass.
#
#     Does some precalculations (like getting list of view methods from
#     the class) to avoid calculating them for each view class instance.
#     """
#
#     def __init__(cls, classname: str, bases: tuple, fields: dict[str, Any]):
#         super().__init__(classname, bases, fields)
#
#         # Gather exposed views
#         cls._urls = []
#         cls._default_view = None
#
#         for p in dir(cls):
#             attr = getattr(cls, p)
#
#             if hasattr(attr, "_urls"):
#                 # Collect methods
#                 for url, methods in attr._urls:
#                     cls._urls.append((url, p, methods))
#
#                     if url == "/":
#                         cls._default_view = p
#
#                         # Wrap views
#                         # setattr(cls, p, _wrap_view(attr))
#
#
# class ModuleComponent:
#     """A component that provide new functions for a :class:`Module`"""
#
#     name: str = None
#
#     def __init__(self, name=None):
#         if name is not None:
#             self.name = name
#
#         if self.name is None:
#             raise ValueError("A module component must have a name")
#
#     def init_module(self, module):
#         self.module = module
#         self.init()
#
#     def init(self, *args, **kwargs):
#         """Implements this in components."""
#
#     def get_actions(self):
#         return []
#
#
# class Module(metaclass=ModuleMeta):
#     id: str = ""
#     endpoint: str = ""
#     label: str = ""
#     managed_class: type | None = None
#     # list_view = None
#     list_view_columns: list[dict[str, Any]] = []
#     single_view = None
#     components: tuple = ()
#
#     # class based views. If not provided will be automaticaly created from
#     # EntityView etc defined below
#     base_template = "base.html"
#     view_cls = EntityView
#     edit_cls = EntityEdit
#     create_cls = EntityCreate
#     delete_cls = EntityDelete
#     json_search_cls = JSONWhooshSearch
#     JSON2_SEARCH_LENGTH = 50
#
#     # form_class. Used when view_cls/edit_cls are not provided
#     edit_form_class = None
#     view_form_class = None  # by default, same as edit_form_class
#
#     url = None
#     name = None
#     view_new_save_and_add = False  # show 'save and add new' button in /new form
#     static_folder = None
#     view_template = None
#     view_options = None
#     related_views: list[RelatedView] = []
#     blueprint = None
#     search_criterions = (
#         search.TextSearchCriterion("name", attributes=("name", "nom")),
#     )
#     # used mostly to change datatable search_label
#     tableview_options = {}  # type: ignore
#     _urls: list[tuple] = []
#
#     def __init__(self):
#         if self.id is None:
#             self.id = self.managed_class.__name__.lower()
#
#         # If endpoint name is not provided, get it from the class name
#         if self.endpoint == "":
#             class_name = self.__class__.__name__
#             if class_name.endswith("Module"):
#                 class_name = class_name[0 : -len("Module")]
#             self.endpoint = class_name.lower()
#
#         if self.label == "":
#             self.label = make_label(self.endpoint)
#
#         # If name is not provided, use capitalized endpoint name
#         if self.name == "":
#             self.name = self._prettify_name(self.__class__.__name__)
#
#         if self.view_options is None:
#             self.view_options = {}
#
#         # self.single_view = make_single_view(self.edit_form_class,
#         #                                     view_template=self.view_template,
#         #                                     **self.view_options)
#         if self.view_form_class is None:
#             self.view_form_class = self.edit_form_class
#
#         # init class based views
#         kw = {
#             "Model": self.managed_class,
#             "pk": "entity_id",
#             "module": self,
#             "base_template": self.base_template,
#         }
#         self._setup_view(
#             "/<int:entity_id>",
#             "entity_view",
#             self.view_cls,
#             Form=self.view_form_class,
#             **kw,
#         )
#         view_endpoint = f"{self.endpoint}.entity_view"
#
#         self._setup_view(
#             "/<int:entity_id>/edit",
#             "entity_edit",
#             self.edit_cls,
#             Form=self.edit_form_class,
#             view_endpoint=view_endpoint,
#             **kw,
#         )
#
#         self._setup_view(
#             "/new",
#             "entity_new",
#             self.create_cls,
#             Form=self.edit_form_class,
#             chain_create_allowed=self.view_new_save_and_add,
#             view_endpoint=view_endpoint,
#             **kw,
#         )
#
#         self._setup_view(
#             "/<int:entity_id>/delete",
#             "entity_delete",
#             self.delete_cls,
#             Form=self.edit_form_class,
#             view_endpoint=view_endpoint,
#             **kw,
#         )
#
#         self._setup_view("/json", "list_json", ListJson, module=self)
#
#         self._setup_view(
#             "/json_search",
#             "json_search",
#             self.json_search_cls,
#             Model=self.managed_class,
#         )
#
#         self.init_related_views()
#
#         # copy criterions instances; without that they may be shared by
#         # subclasses
#         self.search_criterions = copy.deepcopy(self.__class__.search_criterions)
#
#         for sc in self.search_criterions:
#             sc.model = self.managed_class
#
#         self.__components = {}
#         for component in self.components:
#             component.init_module(self)
#             self.__components[component.name] = component
#
#     def get_component(self, name):
#         return self.__components.get(name)
#
#     def _setup_view(self, url: str, attr: str, cls: Any, *args, **kwargs):
#         """Register class based views."""
#         view = cls.as_view(attr, *args, **kwargs)
#         setattr(self, attr, view)
#         self._urls.append((url, attr, view.methods))
#
#     def init_related_views(self):
#         related_views = []
#         for view in self.related_views:
#             if not isinstance(view, RelatedView):
#                 view = DefaultRelatedView(*view)
#             related_views.append(view)
#         self.related_views = related_views
#
#     @property
#     def action_category(self) -> str:
#         return f"module:{self.endpoint}"
#
#     def get_grouped_actions(self) -> OrderedDict:
#         items = actions.for_category(self.action_category)
#         groups = OrderedDict()
#         for action in items:
#             groups.setdefault(action.group, []).append(action)
#
#         return groups
#
#     def register_actions(self):
#         ACTIONS = [
#             ModuleAction(
#                 self,
#                 "entity",
#                 "create",
#                 title=_l("Create New"),
#                 icon=FAIcon("plus"),
#                 endpoint=Endpoint(f"{self.endpoint}.entity_new"),
#                 button="default",
#             )
#         ]
#         for component in self.components:
#             ACTIONS.extend(component.get_actions())
#
#         actions.register(*ACTIONS)
#
#     def create_blueprint(self, crud_app: CRUDApp) -> Blueprint:
#         """Create a Flask blueprint for this module."""
#         # Store admin instance
#         self.crud_app = crud_app
#         self.app = crud_app.app
#
#         # If url is not provided, generate it from endpoint name
#         if self.url is None:
#             self.url = f"{self.crud_app.url}/{self.endpoint}"
#         elif not self.url.startswith("/"):
#             self.url = f"{self.crud_app.url}/{self.url}"
#
#         # Create blueprint and register rules
#         self.blueprint = Blueprint(self.endpoint, __name__, url_prefix=self.url)
#
#         for url, name, methods in self._urls:
#             self.blueprint.add_url_rule(url, name, getattr(self, name), methods=methods)
#
#         # run default_view decorator
#         default_view(self.blueprint, self.managed_class, id_attr="entity_id")(
#             self.entity_view
#         )
#
#         # delay registration of our breadcrumbs to when registered on app; thus
#         # 'parents' blueprint can register theirs befores ours
#         self.blueprint.record_once(self._setup_breadcrumb_preprocessors)
#
#         return self.blueprint
#
#     def _setup_breadcrumb_preprocessors(self, state: BlueprintSetupState):
#         self.blueprint.url_value_preprocessor(self._add_breadcrumb)
#
#     def _add_breadcrumb(self, endpoint: str, values: dict[Any, Any]):
#         g.breadcrumb.append(
#             BreadcrumbItem(label=self.label, url=Endpoint(".list_view"))
#         )
#
#     @property
#     def base_query(self) -> EntityQuery:
#         """Return a query instance for :attr:`managed_class`."""
#         return self.managed_class.query
#
#     @property
#     def read_query(self):
#         """Return a query instance for :attr:`managed_class` filtering on
#         `READ` permission."""
#         return self.base_query.with_permission(READ)
#
#     @property
#     def listing_query(self) -> EntityQuery:
#         """Like `read_query`, but can be made lightweight with only columns and
#         joins of interest.
#
#         `read_query` can be used with exports for example, with lot more
#         columns (generallly it means more joins).
#         """
#         return self.base_query.with_permission(READ)
#
#     def query(self, request: Request):
#         """Return filtered query based on request args."""
#         args = request.args
#         search = args.get("sSearch", "").replace("%", "").lower()
#         query = self.read_query.distinct()
#
#         for crit in self.search_criterions:
#             query = crit.filter(query, self, request, search)
#
#         return query
#
#     def list_query(self, request: Request) -> EntityQuery:
#         """Return a filtered query based on request args, for listings.
#
#         Like `query`, but subclasses can modify it to remove costly
#         joined loads for example.
#         """
#         args = request.args
#         search = args.get("sSearch", "").replace("%", "").lower()
#         query = self.listing_query
#         query = query.distinct()
#
#         for crit in self.search_criterions:
#             query = crit.filter(query, self, request, search)
#
#         return query
#
#     def ordered_query(
#         self, request: Request, query: EntityQuery | None = None
#     ) -> EntityQuery:
#         """Order query according to request args.
#
#         If query is None, the query is generated according to request
#         args with self.query(request)
#         """
#         if query is None:
#             query = self.query(request)
#
#         engine = query.session.get_bind(self.managed_class.__mapper__)
#         args = request.args
#         sort_col = int(args.get("iSortCol_0", 1))
#         sort_dir = args.get("sSortDir_0", "asc")
#         sort_col_def = self.list_view_columns[sort_col]
#         sort_col_name = sort_col_def["name"]
#         rel_sort_names = sort_col_def.get("sort_on", (sort_col_name,))
#         sort_cols = []
#
#         for rel_col in rel_sort_names:
#             sort_col = getattr(self.managed_class, rel_col)
#             if hasattr(sort_col, "property") and isinstance(
#                 sort_col.property, orm.properties.RelationshipProperty
#             ):
#                 # this is a related model: find attribute to filter on
#                 query = query.outerjoin(sort_col_name, aliased=True)
#
#                 rel_model = sort_col.property.mapper.class_
#                 default_sort_name = "name"
#                 if issubclass(rel_model, BaseVocabulary):
#                     default_sort_name = "label"
#
#                 rel_sort_name = sort_col_def.get("relationship_sort_on", None)
#                 if rel_sort_name is None:
#                     rel_sort_name = sort_col_def.get("sort_on", default_sort_name)
#                 sort_col = getattr(rel_model, rel_sort_name, None)
#
#             # XXX: Big hack, date are sorted in reverse order by default
#             if isinstance(sort_col, (Date, DateTime)):
#                 sort_dir = "asc" if sort_dir == "desc" else "desc"
#
#             elif isinstance(sort_col, sa.types.String) or (
#                 hasattr(sort_col, "property")
#                 and isinstance(sort_col.property.columns[0].type, sa.types.String)
#             ):
#                 sort_col = func.lower(sort_col)
#
#             if sort_col is not None:
#                 # FIXME
#                 with contextlib.suppress(Exception):
#                     direction = desc if sort_dir == "desc" else asc
#                     sort_col = direction(sort_col)
#
#                 # sqlite does not support 'NULLS FIRST|LAST' in ORDER BY
#                 # clauses
#                 if engine.name != "sqlite":
#                     nullsorder = nullslast if sort_dir == "desc" else nullsfirst
#                     # FIXME: should not raise an exception
#                     with contextlib.suppress(Exception):
#                         sort_col = nullsorder(sort_col)
#
#                 sort_cols.append(sort_col)
#
#         if sort_cols:
#             # FIXME
#             with contextlib.suppress(Exception):
#                 query = query.order_by(*sort_cols)
#
#         query.reset_joinpoint()
#         return query
#
#     #
#     # Exposed views
#     #
#     @expose("/")
#     def list_view(self) -> str:
#         actions.context["module"] = self
#         table_view = AjaxMainTableView(
#             name=self.managed_class.__name__.lower(),
#             columns=self.list_view_columns,
#             ajax_source=url_for(".list_json"),
#             search_criterions=self.search_criterions,
#             options=self.tableview_options,
#         )
#         rendered_table = table_view.render()
#
#         ctx = {
#             "rendered_table": rendered_table,
#             "module": self,
#             "base_template": self.base_template,
#         }
#         return render_template("default/list_view.html", **ctx)
#
#     def list_json2_query_all(self, q):
#         """Implements the search query for the list_json2 endpoint.
#
#         May be re-defined by a Module subclass in order to customize
#         the search results.
#
#         - Return: a list of results (not json) with an 'id' and a
#           'text' (that will be displayed in the select2).
#         """
#         cls = self.managed_class
#         query = db.session.query(cls.id, cls.name)
#         query = (
#             query.filter(cls.name.ilike(f"%{q}%"))
#             .distinct()
#             .order_by(cls.name)
#             .limit(self.JSON2_SEARCH_LENGTH)
#         )
#         results = query.all()
#         results = [{"id": r[0], "text": r[1]} for r in results]
#         return results
#
#     @expose("/json2")
#     def list_json2(self):
#         """Other JSON endpoint, this time used for filling select boxes
#         dynamically.
#
#         You can write your own search method in list_json2_query_all,
#         that returns a list of results (not json).
#         """
#         args = request.args
#
#         q = args.get("q", "").replace("%", " ")
#         if not q or len(q) < 2:
#             raise BadRequest
#
#         results = self.list_json2_query_all(q)
#         return {"results": results}
#
#     #
#     # Utils
#     #
#     def is_current(self):
#         return request.path.startswith(self.url)
#
#     @staticmethod
#     def _prettify_name(name: str) -> str:
#         """Prettify class name by splitting name by capital characters.
#
#         So, 'MySuperClass' will look like 'My Super Class'
#
#         `name`
#           String to prettify
#         """
#         return re.sub(r"(?<=.)([A-Z])", r" \1", name)
#
#
# class RelatedView:
#     """A base class for related views."""
#
#     def render(self, entity):
#         """Return a dict with keys 'label', 'attr_name', 'rendered', 'size',
#         'show_empty', 'default_collapsed'."""
#         raise NotImplementedError
#
#
# class DefaultRelatedView(RelatedView):
#     """Default view used by Module for items directly related to entity."""
#
#     def __init__(self, label, attr, column_names, options=None, show_empty=False):
#         self.label = label
#         self.attr = attr
#         self.show_empty = show_empty
#         self.column_names = column_names
#         self.options = {}
#
#         if options is not None:
#             self.options.update(options)
#
#     def render(self, entity):
#         view = RelatedTableView(self.column_names, self.options)
#         related_entities = getattr(entity, self.attr)
#         return {
#             "label": self.label,
#             "attr_name": self.attr,
#             "rendered": view.render(related_entities, related_to=entity),
#             "show_empty": self.show_empty,
#             "size": len(related_entities),
#         }
#
#
# # TODO: rename to CRMApp ?
# class CRUDApp:
#     modules: Collection[Module]
#
#     def __init__(self, app: Application, modules: None = None, name: None = None):
#         if name is None:
#             name = self.__class__.__module__
#             modules_signature = ",".join(str(module.id) for module in self.modules)
#             name = f"{name}-{modules_signature}"
#
#         self.name = name
#         self.app = app
#         app.extensions[name] = self
#
#         if modules:
#             self.modules = modules
#
#         for module in self.modules:
#             self.add_module(module)
#
#     def get_module(self, module_id):
#         for m in self.modules:
#             if m.id == module_id:
#                 return m
#
#         return None
#
#     def add_module(self, module: Module):
#         self.app.register_blueprint(self.create_blueprint(module))
#         module.register_actions()
#
#     def create_blueprint(self, module: Module) -> Blueprint:
#         return module.create_blueprint(self)



from __future__ import annotations

from .fields import *  # noqa
from .filters import *  # noqa
from .form import *  # noqa
from .permissions import FormPermissions
from .validators import *  # noqa
from .widgets import *  # noqa



from __future__ import annotations

import logging
import sys

from flask import Flask
from loguru import logger

# Note: prior `patch_logger` feature is now replaced by a simple call to
# loguru, using loguru.trace(). So, by defaut
#
#


def init_logging(app: Flask) -> None:
    run_from_cli = app.config.get("RUN_FROM_CLI")
    if run_from_cli and not app.debug:
        print("Running from CLI, skipping logging init (use --debug to enable)")
        logger.remove()
        intercept_logging_library(app)
    else:
        init_loguru(app)


class InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        logger_opt = logger.opt(depth=6, exception=record.exc_info)
        logger_opt.log(record.levelno, record.getMessage())


def intercept_logging_library(app: Flask) -> None:
    logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)
    app.logger.handlers = []
    app.logger.addHandler(InterceptHandler())


def init_loguru(app: Flask):
    # register loguru as (sole) handler
    level = app.config.get("LOG_LEVEL")
    if not level:
        if app.debug:
            level = "DEBUG"
        else:
            level = "INFO"

    set_loguru_config(level)
    intercept_logging_library(app)
    logger.debug("Loguru initialized")


def set_loguru_config(level: str = "DEBUG"):
    logger.remove()
    logger.add(
        sys.stderr,
        level=level,
        backtrace=True,
        diagnose=True,
        enqueue=True,
        catch=True,
    )



"""Objects to schema adapters."""

from __future__ import annotations

from abc import ABC, abstractmethod
from operator import attrgetter
from typing import Any

import sqlalchemy as sa
from loguru import logger
from sqlalchemy.orm.session import Session
from whoosh.fields import ID, TEXT, Schema

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.models import Model

from .schema import accent_folder

__all__ = ["SAAdapter", "SchemaAdapter"]


class SchemaAdapter(ABC):
    """Abstract base class for objects to schema adapter.

    The purpose of adapters is that given an object they return kwargs
    for document.
    """

    def __init__(self, model_class, schema):
        """
        :param:model_class: class of objects instances to be adapted
        :param:schema: :class:`whoosh.fields.Schema` instance
        """

    @staticmethod
    def can_adapt(obj_cls):
        """Return True if this class can adapt objects of class `obj_cls`."""
        raise NotImplementedError

    @abstractmethod
    def retrieve(self, pk, **data):
        """Returns an object instance given its identifier and optional data
        kwargs.

        :param pk: the primary key to retrieve the object

        :returns: None if not object instance can be returned
        """
        raise NotImplementedError

    @abstractmethod
    def get_document(self, obj):
        raise NotImplementedError


class SAAdapter(SchemaAdapter):
    """Adapts an sqlalchemy model.

    A column on the model is indexed if `"searchable"` is set to `True` in
    its `info` field (see :meth:`sqlalchemy.schema.Column.__init__`)

    Additionnaly info can also contains `index_to`. It's expected to be an
    iterable of field names or tuple of (field name, whoosh field
    instance). The default the whoosh field is a
    :class:`whoosh.fields.TEXT` instance.

    When `index_to` is used the column name will *not* be indexed on the same
    field name unless specified in `index_to`.

    If multiple field names are passed to `index_to` the values will be
    concatenated with a space.

    `index_to` is useful for::

    * index many attributes to on field names, for exemple index `title`
    and `description` to a `_full_text` field
    * define new fields on schema
    """

    doc_attrs: dict[str, Any]

    def __init__(self, model_class: type[Model], schema: Schema):
        """
        :param:model_class: a sqlalchemy model class
        :param:schema: :class:`whoosh.fields.Schema` instance
        """
        assert issubclass(model_class, db.Model)

        self.model_class = model_class
        self.indexable = getattr(model_class, "__indexable__", False)
        self.index_to = self.get_index_to(model_class)
        self.doc_attrs = {}
        if self.indexable:
            self._build_doc_attrs(model_class, schema)

    @staticmethod
    def can_adapt(obj_cls: Any) -> bool:
        return issubclass(obj_cls, db.Model)

    def get_index_to(self, model_class: type[Model]) -> tuple:
        result = []
        classes = model_class.mro()
        for cls in classes:
            if hasattr(cls, "__index_to__"):
                result += cls.__index_to__
        return tuple(result)

    def _build_doc_attrs(self, model_class: type[Model], schema: Schema):
        mapper = sa.inspect(model_class)

        args = self.doc_attrs
        # Any field not in schema will be stored here.
        # After all field have been discovered, we add the missing ones.
        field_definitions = {}

        def setup_field(attr_name: str, field_name: tuple[str, type | ID] | str):
            field_def = False
            if not isinstance(field_name, str):
                field_name, field_def = field_name

            if field_name not in schema and (
                field_name not in field_definitions
                or field_definitions[field_name] is False
            ):
                field_definitions[field_name] = field_def

            # attrgetter offers dotted name support. Useful for attributes on
            # related objects.
            args.setdefault(field_name, {})[attr_name] = attrgetter(attr_name)

        # model level definitions
        for name, field_names in self.index_to:
            if isinstance(field_names, str):
                field_names = (field_names,)
            for field_name in field_names:
                setup_field(name, field_name)

        # per column definitions
        for col in mapper.columns:
            name = col.name
            info = col.info

            if not info.get("searchable"):
                continue

            index_to = info.get("index_to", (name,))
            if isinstance(index_to, str):
                index_to = (index_to,)

            for field_name in index_to:
                setup_field(name, field_name)

        # add missing fields to schema
        for field_name, field_def in field_definitions.items():
            if field_name in schema:
                continue

            if field_def is False:
                field_def = TEXT(stored=True, analyzer=accent_folder)

            logger.debug(
                "Adding field to schema:\n  Model: {model}\n"
                '  Field: "{field_name}" {field_def}',
                model=model_class._object_type(),
                field_name=field_name,
                field_def=field_def,
            )
            schema.add(field_name, field_def)

    def retrieve(self, pk: int, _session: Session | None = None, **data: Any) -> Entity:
        if _session is None:
            _session = db.session()
        return _session.query(self.model_class).get(pk)

    def get_document(self, obj: Model) -> dict[str, Any]:
        result: dict[str, Any] = {}
        if not self.indexable:
            return result

        # Cache because the same attribute may be needed by many fields, i.e
        # "title" on "title" field and "full_text" field for example.
        cached = {}
        # Negative cache. Might be used especially with dotted names.
        missed = set()

        for field_name, attrs in self.doc_attrs.items():
            values = []

            for a, getter in attrs.items():
                if a in missed:
                    continue

                if a not in cached:
                    try:
                        val = getter(obj)
                    except AttributeError:
                        missed.add(a)
                        continue
                    else:
                        cached[a] = val

                val = cached[a]
                if val is not None:
                    if isinstance(val, (list, tuple)):
                        val = " ".join(val).strip()
                    values.append(val)

            values = [v for v in values if v]

            if values:
                if len(values) == 1:
                    result[field_name] = values[0]
                else:
                    result[field_name] = " ".join(values)

        return result



""""""

from __future__ import annotations

import contextlib

from flask import render_template

from abilian.i18n import _
from abilian.services.auth.models import LoginSession
from abilian.web.admin import AdminPanel

from .geoip.ip_country_code import ip_to_country_code


class LoginSessionsPanel(AdminPanel):
    id = "login_sessions"
    label = "Session log"
    icon = "log-in"

    def get(self) -> str:
        sessions = LoginSession.query.order_by(LoginSession.id.desc()).limit(50).all()
        unknown_country = _("Country unknown")

        for session in sessions:
            if session.ip_address:
                # if several IPs, only use last IP in the list, most likely the
                # public address
                ip_address = session.ip_address.rpartition(",")[-1]
                session.country = ip_to_country_code(ip_address) or unknown_country

        ctx = {"sessions": sessions}
        return render_template("admin/login_sessions.html", **ctx)



from flask import current_app, g, session

from abilian.core.entities import Entity

# Not used in UI, currently.


def add_to_recent_items(entity, type="ignored"):
    if not isinstance(entity, Entity):
        return
    object_type = entity.object_type
    url = current_app.default_view.url_for(entity)
    if not hasattr(g, "recent_items"):
        g.recent_items = []
    g.recent_items.insert(0, {"type": object_type, "name": entity.name, "url": url})

    s = set()
    new_recent_items = []
    for item in g.recent_items:
        item_url = item["url"]
        if item_url in s:
            continue
        s.add(item_url)
        new_recent_items.append(item)
    if len(new_recent_items) > 5:
        del new_recent_items[5:]
    session["recent_items"] = g.recent_items = new_recent_items



"""Forum views."""

from __future__ import annotations

from datetime import datetime

from flask import g
from flask_babel import format_date

from abilian.i18n import _l
from abilian.sbe.apps.communities.security import is_manager
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.wiki.models import WikiPage
from abilian.services.viewtracker import viewtracker


def object_viewers(entity: Document | WikiPage) -> list:
    if is_manager():
        views = viewtracker.get_views(entity=entity)
        community_members_id = [
            user.id for user in g.community.members if user.id != entity.creator.id
        ]
        viewers = []
        for view in views:
            if view.user_id in set(community_members_id):
                viewers.append(
                    {"user": view.user, "viewed_at": view.hits[-1].viewed_at}
                )
        return viewers
    return []


def activity_time_format(time: datetime, now: datetime | None = None) -> str:
    if not time:
        return ""

    if not now:
        now = datetime.utcnow()
    time_delta = now - time
    month_abbreviation = format_date(time, "MMM")
    days, hours, minutes, seconds = (
        time_delta.days,
        time_delta.seconds // 3600,
        time_delta.seconds // 60,
        time_delta.seconds,
    )

    if days == 0 and hours == 0 and minutes == 0:
        return f"{seconds}{_l('s')}"

    if days == 0 and hours == 0:
        return f"{minutes}{_l('m')}"

    if days == 0:
        return f"{hours}{_l('h')}"

    if days < 30:
        return f"{days}{_l('d')}"

    if time.year == now.year:
        return f"{month_abbreviation} {time.day}"

    return f"{month_abbreviation} {time.year!s}"



from __future__ import annotations

import dramatiq_abort.backends
import redis
from dramatiq.rate_limits import BucketRateLimiter
from dramatiq.rate_limits.backends import RedisBackend as RateLimitRedisBackend
from dramatiq_abort import Abortable
from loguru import logger

from .cli import scheduler
from .singleton import dramatiq

DEFAULT_DRAMATIQ_ABORT_REDIS_URL = "redis://localhost:6379/0"
DEFAULT_DRAMATIQ_RATE_LIMIT_REDIS_URL = "redis://localhost:6379/0"
RATE_LIMITER = []


# with DISTRIBUTED_MUTEX.acquire():
#         time.sleep(1)
def init_dramatiq_engine(app) -> None:
    if dramatiq.app is not None:
        # ensure only one initialization, especially during tests
        return
    logger.debug("Setting up Dramatiq")
    dramatiq.init_app(app)
    _setup_rate_limiter(app)
    _add_dramatiq_abortable(app)
    _register_scheduler(app)
    _print_dramatiq_config()


def _setup_rate_limiter(app):
    logger.debug("Add dramatiq rate limiter")
    redis_client = _rate_limiter_redis_client(app)
    backend = RateLimitRedisBackend(client=redis_client)
    # rate limit is 6/min
    limiter = BucketRateLimiter(backend, "mail-limit", limit=6, bucket=60_000)
    RATE_LIMITER.append(limiter)


def _rate_limiter_redis_client(app) -> redis.Redis:
    redis_url = app.config.get("DRAMATIQ_RATE_LIMIT_REDIS_URL")
    if not redis_url:
        redis_url = app.config.get("DRAMATIQ_BROKER_URL")
    if not redis_url:
        redis_url = DEFAULT_DRAMATIQ_RATE_LIMIT_REDIS_URL
    return redis.Redis.from_url(redis_url)


def _register_scheduler(app) -> None:
    logger.debug("Register scheduler")
    app.cli.add_command(scheduler)


def _abortable_redis_client(app) -> redis.Redis:
    redis_url = app.config.get("DRAMATIQ_ABORT_REDIS_URL")
    if not redis_url:
        redis_url = app.config.get("DRAMATIQ_BROKER_URL")
    if not redis_url:
        redis_url = DEFAULT_DRAMATIQ_ABORT_REDIS_URL
    return redis.Redis.from_url(redis_url)


def _add_dramatiq_abortable(app) -> None:
    """Configure abort feature of dramatiq tasks.

    The dramatiq-abort package provides a middleware that can be used to
    abort running actors by message id. Here s how you might set it up:

    @dramatiq.actor
    def a_long_running_task():
        ...

    message = a_long_running_task.send()
    abort(message.message_id)

    abort(message_id, mode=AbortMode.CANCEL)
    abort(message.message_id, mode=AbortMode.ABORT, abort_timeout=2000)
    """
    logger.debug("Add dramatiq abortable")
    redis_client = _abortable_redis_client(app)
    backend = dramatiq_abort.backends.RedisBackend(client=redis_client)
    abortable = Abortable(backend=backend)
    dramatiq.broker.add_middleware(abortable)


def _print_dramatiq_config() -> None:
    """Show tasks broker configuration."""
    broker = dramatiq.broker
    messages = []
    messages.append("Dramatiq broker in app:")
    messages.append(f"{broker=}")
    messages.append("broker middlewares:")
    middlewares = ", ".join([str(mw.__class__.__name__) for mw in broker.middleware])
    messages.append(middlewares)
    messages.append("broker.get_declared_queues():")
    messages.append(str(broker.get_declared_queues()))
    messages.append("broker.get_declared_actors():")
    messages.append(str(broker.get_declared_actors()))
    for msg in messages:
        logger.debug(msg)



"""BaseMixin class."""

from __future__ import annotations

import json
from datetime import datetime
from typing import Any

from sqlalchemy.orm.util import class_mapper

from .base import IdMixin, TimestampedMixin
from .owned import OwnedMixin


class BaseMixin(IdMixin, TimestampedMixin, OwnedMixin):
    name: str

    def __init__(self):
        OwnedMixin.__init__(self)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"instance at 0x{id(self):x} name={self.name!r} id={self.id}>"
        )

    def __str__(self):
        return self.name or ""

    @property
    def column_names(self) -> list[str]:
        return [col.name for col in class_mapper(self.__class__).persist_selectable.c]

    def to_dict(self) -> dict[str, Any]:
        if hasattr(self, "__exportable__"):
            exported = self.__exportable__ + ["id"]
        else:
            exported = self.column_names
        result: dict[str, Any] = {}
        for key in exported:
            value = getattr(self, key)
            if isinstance(value, datetime):
                value = value.isoformat()
            result[key] = value
        return result

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    def _icon(self, size: int = 12) -> str:
        class_name = self.__class__.__name__.lower()
        return f"/static/icons/{class_name}-{size}.png"

    # FIXME: we can do better than that
    @property
    def _name(self):
        if hasattr(self, "title"):
            return self.title
        elif hasattr(self, "name"):
            return self.name
        else:
            raise NotImplementedError



""""""

from __future__ import annotations

from flask import current_app, flash, request
from flask.signals import request_started
from flask_wtf.csrf import CSRFError, CSRFProtect
from markupsafe import Markup
from werkzeug.exceptions import BadRequest

from abilian.core.util import unwrap
from abilian.i18n import _l

wtf_csrf = CSRFProtect()


class AbilianCsrf:
    """CSRF error handler, that allows supporting views to gracefully report
    error instead of a plain 400 error.

    views supporting this must
    """

    #: for views that gracefully support csrf errors, this message can be
    #: displayed to user. It can be changed if you have a better one for your
    #: users.
    csrf_failed_message = _l(
        "Security informations are missing or expired. "
        "This may happen if you have opened the form for a long time. "
        "<br /><br />"
        "Please try to resubmit the form."
    )

    def init_app(self, app):
        if "csrf" not in app.extensions:
            raise RuntimeError(
                'Please install flask_wtf.csrf.CSRFProtect() as "csrf" in '
                "extensions before AbilianCsrf()"
            )

        # FIXME
        # app.extensions["csrf"].error_handler(self.csrf_error_handler)
        # Note: deprecated since flask-wtf 0.14

        # see https://flask-wtf.readthedocs.io/en/1.2.x/changes/#version-0-14
        #     https://github.com/wtforms/flask-wtf/pull/264

        wtf_csrf.init_app(app)

        app.register_error_handler(CSRFError, self.csrf_error_handler)

        app.extensions["csrf-handler"] = self
        request_started.connect(self.request_started, sender=app)
        app.before_request(self.before_request)

    def flash_csrf_failed_message(self):
        flash(Markup(self.csrf_failed_message), "error")

    def request_started(self, app):
        request.csrf_failed = False

    def csrf_error_handler(self, reason):
        request.csrf_failed = reason

    def before_request(self):
        req = unwrap(request)
        failed = req.csrf_failed

        if not failed:
            return

        rule = req.url_rule
        view = current_app.view_functions[rule.endpoint]
        if getattr(view, "csrf_support_graceful_failure", False):
            # view can handle it nicely for the user
            return

        if hasattr(view, "view_class") and getattr(
            view.view_class, "csrf_support_graceful_failure", False
        ):
            return

        raise BadRequest(failed)


abilian_csrf = AbilianCsrf()



from __future__ import annotations

from datetime import datetime

from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.types import DateTime, Integer

from abilian.core.entities import Entity, db
from abilian.core.models.subjects import User


# TODO: remove duplicate
def _default_from(column):
    """Helper for default and onupdates parameters in a Column definitions.

    Returns a `context-sensitive default function
    <http://docs.sqlalchemy.org/en/rel_0_8/core/defaults.html#context-
    sensitive-default-functions>`_ to set value from another column.
    """

    def _default_value(context):
        return context.current_parameters[column]

    return _default_value


class View(db.Model):
    __tablename__ = "view"

    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)

    #: viewed entity id
    entity_id = Column(Integer, default=_default_from("_fk_entity_id"), nullable=False)
    _fk_entity_id = Column(Integer, ForeignKey(Entity.id, ondelete="SET NULL"))
    entity = relationship(Entity, foreign_keys=_fk_entity_id)

    #: user id
    user_id = Column(Integer, ForeignKey(User.id), nullable=False)
    user = relationship(User, foreign_keys=user_id)

    hits = relationship("Hit", backref="view", order_by="Hit.viewed_at", lazy="dynamic")


class Hit(db.Model):
    __tablename__ = "hit"

    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    view_id = Column(Integer, ForeignKey(View.id))

    #: time of view
    viewed_at = Column(DateTime, default=datetime.utcnow, nullable=True)



import functools
from collections.abc import Callable
from pathlib import Path

from filelock import FileLock, Timeout

from .exceptions import ConversionError

LOCK_EXPIRE = 1800  # 30 min, in case many request in //
LOCK_DIR = []
LOCK_FILES = {}


def init_conversion_lock_dir(instance_path: str) -> None:
    lock_dir = Path(instance_path) / "lock"
    if not lock_dir.exists():
        lock_dir.mkdir(parents=True)
    LOCK_DIR.append(lock_dir)


def get_lock_file(name: str) -> FileLock:
    if lock := LOCK_FILES.get(name):
        return lock
    lock = FileLock(LOCK_DIR[0] / f"{name}.lock", timeout=LOCK_EXPIRE)
    LOCK_FILES[name] = lock
    return lock


def acquire_lock(name: str) -> Callable:
    """Ensure the decorated function is alone to run by using a lock file."""

    def locked(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            lock = get_lock_file(name)
            try:
                lock.acquire()
                return func(*args, **kwargs)
            except Timeout as e:
                raise ConversionError from e
            finally:
                lock.release()

        return wrapper

    return locked



"""Base Flask application class, used by tests or to be extended in real
applications."""

from __future__ import annotations

from functools import partial

import sqlalchemy as sa
from flask import Flask, g, render_template
from flask.globals import request_ctx
from loguru import logger

from abilian.core import extensions

db = extensions.db


class ErrorManagerMixin(Flask):
    def setup_logging(self):
        # Force flask to create application logger before logging
        # configuration; else, flask will overwrite our settings
        assert self.logger

    def handle_user_exception(self, e):
        # If session.transaction._parent is None, then exception has occured in
        # after_commit(): doing a rollback() raises an error and would hide
        # actual error.
        session = db.session()
        if session.is_active and session.transaction._parent is not None:
            # Inconditionally forget all DB changes, and ensure clean session
            # during exception handling.
            session.rollback()
        else:
            self._remove_session_save_objects()

        return super().handle_user_exception(e)

    def handle_exception(self, e):
        session = db.session()
        if not session.is_active:
            # Something happened in error handlers and session is not usable
            # anymore.
            self._remove_session_save_objects()

        return super().handle_exception(e)

    def _remove_session_save_objects(self):
        """Used during exception handling in case we need to remove() session:

        keep instances and merge them in the new session.
        """
        if self.testing:
            return
        # Before destroying the session, get all instances to be attached to the
        # new session. Without this, we get DetachedInstance errors, like when
        # tryin to get user's attribute in the error page...
        old_session = db.session()
        g_objs = []
        for key in iter(g):
            obj = getattr(g, key)
            if isinstance(obj, db.Model) and sa.orm.object_session(obj) in (
                None,
                old_session,
            ):
                g_objs.append((key, obj, obj in old_session.dirty))

        db.session.remove()
        session = db.session()

        for key, obj, load in g_objs:
            # replace obj instance in bad session by new instance in fresh
            # session
            setattr(g, key, session.merge(obj, load=load))

        # refresh `current_user`
        user = getattr(request_ctx, "user", None)
        if user is not None and isinstance(user, db.Model):
            request_ctx.user = session.merge(user, load=load)

    def log_exception(self, exc_info):
        """Log exception only if Sentry is not used (this avoids getting error
        twice in Sentry)."""
        dsn = self.config.get("SENTRY_DSN")
        if not dsn:
            super().log_exception(exc_info)

    def init_sentry(self):
        """Install Sentry handler if config defines 'SENTRY_DSN'."""
        dsn = self.config.get("SENTRY_DSN")
        if not dsn:
            return

        try:
            import sentry_sdk
        except ImportError:
            logger.error(
                'SENTRY_DSN is defined in config but package "sentry-sdk"'
                " is not installed."
            )
            return

        from sentry_sdk.integrations.flask import FlaskIntegration
        from sentry_sdk.integrations.redis import RedisIntegration
        from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

        sentry_sdk.init(
            dsn=dsn,
            integrations=[
                FlaskIntegration(),
                SqlalchemyIntegration(),
                RedisIntegration(),
            ],
        )

    def install_default_handlers(self):
        for http_error_code in (403, 404, 500):
            self.install_default_handler(http_error_code)

    def install_default_handler(self, http_error_code: int):
        """Install a default error handler for `http_error_code`.

        The default error handler renders a template named error404.html
        for http_error_code 404.
        """
        logger.debug(
            "Set Default HTTP error handler for status code {http_error_code}",
            http_error_code=http_error_code,
        )
        handler = partial(self.handle_http_error, http_error_code)
        self.errorhandler(http_error_code)(handler)

    def handle_http_error(self, code, error):
        """Helper that renders `error{code}.html`.

        Convenient way to use it::

           from functools import partial
           handler = partial(app.handle_http_error, code)
           app.errorhandler(code)(handler)
        """
        # 5xx code: error on server side
        if (code // 100) == 5:
            # ensure rollback if needed, else error page may
            # have an error, too, resulting in raw 500 page :-(
            db.session.rollback()

        template = f"error{code:d}.html"
        return render_template(template, error=error), code



"""Entity objects for the Document Management applications.

TODO: move to an independent service / app.
"""

from __future__ import annotations

import itertools
import mimetypes
import threading
import uuid
from collections.abc import Collection, Iterator
from importlib import resources as rso
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
import whoosh.fields as wf
from flask import current_app, json, url_for
from flask_login import current_user

# from loguru import logger
from loguru import logger
from sqlalchemy.event import listen, listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, foreign, relationship, remote
from sqlalchemy.orm.attributes import Event
from sqlalchemy.orm.session import Session
from sqlalchemy.schema import Column, ForeignKey, UniqueConstraint
from sqlalchemy.types import Integer, Text, UnicodeText
from sqlalchemy.util.langhelpers import symbol
from toolz import first
from whoosh.analysis import CharsetFilter, LowercaseFilter, RegexTokenizer
from whoosh.support.charset import accent_map

from abilian.core.entities import Entity, db
from abilian.core.models import NOT_AUDITABLE, SEARCHABLE
from abilian.core.models.blob import Blob
from abilian.core.models.subjects import Group, User
from abilian.core.util import md5
from abilian.services.conversion import converter
from abilian.services.indexing import indexable_role
from abilian.services.security import Admin, Anonymous, InheritSecurity, security

from . import tasks
from .lock import Lock

if TYPE_CHECKING:
    from abilian.sbe.apps.communities.models import Community


# logger = logging.getLogger(__package__)

__all__ = (
    "BaseContent",
    "CmisObject",
    "Document",
    "Folder",
    "PathAndSecurityIndexable",
    "icon_for",
    "icon_url",
    "setup_listener",
)

#: A Whoosh analyzer that folds accents and case.
accent_folder = RegexTokenizer() | LowercaseFilter() | CharsetFilter(accent_map)

ICONS_FOLDER = rso.files("abilian.sbe") / "static" / "fileicons"


def icon_url(filename: str) -> str:
    return url_for("abilian_sbe_static", filename=f"fileicons/{filename}")


def icon_exists(filename: str) -> bool:
    return ICONS_FOLDER.joinpath(filename).is_file()


#
# Domain classes
#
class CmisObject(InheritSecurity, Entity):
    """(Abstract) Base class for CMIS objects."""

    # normally set by communities.models.community_content,
    # but we have to fix a circ.dep to use it
    is_community_content = True

    __tablename__ = "cmisobject"
    __indexable__ = False
    __index_to__ = (
        ("community.id", ("community_id",)),
        #
        ("community.slug", ("community_slug",)),
    )

    _title = Column("title", UnicodeText, nullable=False, default="")
    description = Column(
        UnicodeText,
        nullable=False,
        default="",
        info=SEARCHABLE | {"index_to": ("description", "text")},
    )

    _parent_id = Column(Integer, ForeignKey("cmisobject.id"), nullable=True)

    # no duplicate name in same folder
    __table_args__ = (UniqueConstraint("_parent_id", "title"),)

    # Set in concrete classes
    sbe_type: str = ""

    # Convenience default values
    content_length = 0

    def __init__(self, *args, **kwargs):
        # ensure 'title' prevails over 'name'
        if "title" in kwargs and "name" in kwargs:
            title = kwargs.get("title")
            name = kwargs.get("name")

            if title is None:
                del kwargs["title"]
            elif name != title:
                kwargs["name"] = title

        Entity.__init__(self, *args, **kwargs)

    # title is defined has an hybrid property to allow 2 way sync name <->
    # title
    @hybrid_property
    def title(self: str) -> str:
        return self._title

    @title.setter
    def title(self, title: str) -> None:
        # set title before setting name, so that we don't enter an infinite loop
        # with _cmis_sync_name_title
        self._title = title
        if self.name != title:
            self.name = title

    def clone(
        self, title: str | None = None, parent: Folder | None = None
    ) -> CmisObject:
        if not title:
            title = self.title
        new_obj = self.__class__(title=title, name=title, parent=parent)
        state = vars(self)
        for k, v in state.items():
            if not k.startswith("_") and k not in [
                "uid",
                "id",
                "parent",
                "title",
                "name",
                "path",
                "subfolders",
                "documents",
            ]:
                setattr(new_obj, k, v)
        if self.parent:
            assert self in self.parent.children
        return new_obj

    @property
    def path(self) -> str:
        if self.parent:
            return f"{self.parent.path}/{self.title}"
        else:
            return ""

    @property
    def is_folder(self) -> bool:
        return self.sbe_type == "cmis:folder"

    @property
    def is_document(self) -> bool:
        return self.sbe_type == "cmis:document"

    @property
    def is_root_folder(self) -> bool:
        return self._parent_id is None

    @property
    def community(self) -> Community | None:
        if not self.is_folder:
            return self.parent and self.parent.community

        if self._community:
            return self._community

        return self.parent and self.parent.community


@listens_for(CmisObject.name, "set", propagate=True, active_history=True)
def _cmis_sync_name_title(
    entity: CmisObject, new_value: str, old_value: symbol | str, initiator: Event
) -> str:
    """Synchronize CmisObject name -> title.

    CmisObject.title -> name is done via hybrid_property, avoiding
    infinite loop (since "set" is received before attribute has received
    value)
    """
    if entity.title != new_value:
        entity.title = new_value
    return new_value


class PathAndSecurityIndexable:
    """Mixin for folder and documents indexation."""

    __index_to__ = (
        ("_indexable_parent_ids", ("parent_ids",)),
        ("_indexable_roles_and_users", ("allowed_roles_and_users",)),
    )

    def _iter_to_root(self, skip_self: bool = False) -> Iterator[BaseContent]:
        obj = self if not skip_self else self.parent
        while obj:
            yield obj
            obj = obj.parent

    @property
    def _indexable_parent_ids(self) -> str:
        """Return a string made of ids separated by a slash: "/1/3/4/5", "5"
        being self.parent.id."""
        ids = [str(obj.id) for obj in self._iter_to_root(skip_self=True)]
        return f"/{'/'.join(reversed(ids))}"

    @property
    def _indexable_roles_and_users(self) -> str:
        """Returns a string made of type:id elements, like "user:2 group:1
        user:6"."""
        iter_from_root = reversed(list(self._iter_to_root()))
        if self.parent:
            # skip root folder only on non-root folder!
            next(iter_from_root)
        allowed = {o[0] for o in security.get_role_assignements(next(iter_from_root))}

        for obj in iter_from_root:
            if obj.inherit_security:
                continue
            obj_allowed = {o[0] for o in security.get_role_assignements(obj)}

            if Anonymous in obj_allowed:
                continue

            parent_allowed = allowed
            # pure intersection: users and groups in both are preserved
            allowed = allowed & obj_allowed
            remaining = parent_allowed - obj_allowed
            # find users who can access 'obj' because of their group memberships
            # 1. extends groups in obj_allowed with their actual member list
            extended_allowed = set(
                itertools.chain(
                    *(p.members if isinstance(p, Group) else (p,) for p in obj_allowed)
                )
            )

            # 2. remaining_users are users explicitly listed in parents but not on
            # obj. Are they in a group?
            remaining_users = {o for o in remaining if isinstance(o, User)}
            allowed |= remaining_users & extended_allowed

            # remaining groups: find if some users are eligible
            remaining_groups_members = set(
                itertools.chain(*(p.members for p in remaining if isinstance(p, Group)))
            )
            allowed |= remaining_groups_members - extended_allowed

        # admin role is always granted access
        allowed.add(Admin)
        return " ".join(indexable_role(p) for p in allowed)


class Folder(PathAndSecurityIndexable, CmisObject):
    __tablename__ = "folder"
    sbe_type = "cmis:folder"

    __indexable__ = True

    _indexable_roles_and_users = PathAndSecurityIndexable._indexable_roles_and_users

    parent = relationship(
        "Folder",
        primaryjoin=(lambda: foreign(CmisObject._parent_id) == remote(Folder.id)),
        backref=backref(
            "subfolders",
            lazy="joined",
            order_by="Folder.title",
            cascade="all, delete-orphan",
        ),
    )

    @property
    def icon(self):
        return icon_url("folder.png")

    @property
    def children(self) -> list[BaseContent]:
        return self.subfolders + self.documents

    @property
    def document_count(self) -> int:
        count = len(self.documents)
        for f in self.subfolders:
            count += f.document_count
        return count

    @property
    def depth(self) -> int:
        if self.parent is None:
            return 0

        return self.parent.depth + 1

    def create_subfolder(self, title: str) -> Folder:
        subfolder = Folder(title=title, parent=self)
        assert subfolder in self.children
        return subfolder

    def create_document(self, title: str) -> Document:
        doc = Document(title=title, parent=self)
        assert doc.parent == self
        assert doc in self.children
        return doc

    def get_object_by_path(self, path: str) -> BaseContent | None:
        assert path.startswith("/")
        assert "//" not in path

        if path == "/":
            return self

        path_segments = path[1:].split("/")
        obj = self
        try:
            for name in path_segments[:]:
                obj = first(x for x in obj.children if x.title == name)
            return obj
        except IndexError:
            return None

    def __repr__(self):
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__} "
            f"id={self.id!r} "
            f"name={self.title!r} "
            f"path={self.path!r} "
            f"at 0x{id(self):x}>"
        )

    #
    # Security related methods
    #
    @property
    def filtered_children(self) -> list[BaseContent]:
        return security.filter_with_permission(
            current_user, "read", self.children, inherit=True
        )

    @property
    def filtered_subfolders(self) -> list[Folder]:
        return security.filter_with_permission(
            current_user, "read", self.subfolders, inherit=True
        )

    def get_local_roles_assignments(self):
        local_roles_assignments = security.get_role_assignements(self)
        # local_roles_assignments = sorted(local_roles_assignments,
        #                                  key=lambda u: (u[0].last_name.lower(),
        #                                                 u[0].first_name.lower()))
        return local_roles_assignments

    def get_inherited_roles_assignments(self):
        if self.parent.is_root_folder:
            return []

        roles = self.parent.get_local_roles_assignments()
        inherited_roles = (
            self.parent.get_inherited_roles_assignments()
            if self.parent.inherit_security
            else []
        )

        assignments = set(itertools.chain(roles, inherited_roles))

        def key(x):
            principal = x[0]
            if isinstance(principal, User):
                # Defensive programming here, this shouldn't happen actually
                last_name = principal.last_name or ""
                first_name = principal.first_name or ""
                return last_name.lower(), first_name.lower()
            elif isinstance(principal, Group):
                return principal.name
            else:
                raise TypeError(f"Bad class here: {type(principal)}")

        return sorted(assignments, key=key)

    def members(self) -> Collection[User]:
        local_roles = self.get_local_roles_assignments()
        inherited_roles = (
            self.get_inherited_roles_assignments() if self.inherit_security else []
        )

        def _iter_users(roles):
            for principal, _user in roles:
                if isinstance(principal, User):
                    yield principal
                else:
                    yield from itertools.chain(principal.members, principal.admins)

        members = set(_iter_users(itertools.chain(local_roles, inherited_roles)))
        members = sorted(members, key=lambda u: (u.last_name, u.first_name))
        return members


class BaseContent(CmisObject):
    """A base class for cmisobject with an attached file."""

    __tablename__ = "base_content"
    _content_id = Column(Integer, db.ForeignKey(Blob.id))
    content_blob = relationship(Blob, cascade="all, delete", foreign_keys=[_content_id])

    #: md5 digest (BTW: not sure they should be part of the public API).
    content_digest = Column(Text)

    #: size (in bytes) of the content blob.
    content_length = Column(
        Integer,
        default=0,
        nullable=False,
        server_default=sa.text("0"),
        info={
            "searchable": True,
            "index_to": (("content_length", wf.NUMERIC(stored=True)),),
        },
    )

    #: MIME type of the content stream.
    # TODO: normalize mime type?
    content_type = Column(
        Text,
        default="application/octet-stream",
        info={"searchable": True, "index_to": (("content_type", wf.ID(stored=True)),)},
    )

    @property
    def content(self) -> bytes:
        return self.content_blob.value

    @content.setter
    def content(self, value: bytes) -> None:
        assert isinstance(value, bytes)
        self.content_blob = Blob()
        self.content_blob.value = value
        self.content_length = len(value)

    def set_content(self, content: bytes, content_type: str = "") -> None:
        assert isinstance(content_type, str)

        new_digest = md5(content)
        if new_digest == self.content_digest:
            return

        self.content_digest = new_digest
        self.content = content
        content_type = self.find_content_type(content_type)
        if content_type:
            self.content_type = content_type

    def find_content_type(self, content_type: str = "") -> str:
        """Find possibly more appropriate content_type for this instance.

        If `content_type` is a binary one, try to find a better one
        based on content name so that 'xxx.pdf' is not flagged as
        binary/octet-stream for example
        """
        assert isinstance(content_type, str)

        if content_type not in (
            "",
            "application/octet-stream",
            "binary/octet-stream",
            "application/binary",
            "multipart/octet-stream",
        ):
            return content_type

        # missing or generic content type: try to find something more useful to be
        # able to do preview/indexing/...
        if self.title:
            guessed_content_type = mimetypes.guess_type(self.title, strict=False)[0]
            if (
                guessed_content_type
                and guessed_content_type != "application/vnd.ms-office.activeX"
            ):
                # mimetypes got an update: "random.bin" would be guessed as
                # 'application/vnd.ms-office.activeX'... not so useful in a document
                # repository
                content_type = guessed_content_type

        return content_type

    @property
    def icon(self) -> str:
        return icon_for(self.content_type)


class Document(BaseContent, PathAndSecurityIndexable):
    """A document, in the CMIS sense."""

    __tablename__ = "document"
    __indexable__ = True
    __index_to__ = (("text", ("text",)),)

    _indexable_roles_and_users = PathAndSecurityIndexable._indexable_roles_and_users

    parent = relationship(
        "Folder",
        primaryjoin=(foreign(CmisObject._parent_id) == remote(Folder.id)),
        backref=backref(
            "documents",
            lazy="joined",
            order_by="Document.title",
            cascade="all, delete-orphan",
        ),
    )

    PREVIEW_SIZE = 700

    @property
    def preview_size(self) -> int:
        return self.PREVIEW_SIZE

    def has_preview(self, size: int = 0, index: int = 0) -> bool:
        if size == 0:
            size = self.PREVIEW_SIZE

        return converter.has_image(self.content_digest, self.content_type, index, size)

    @property
    def digest(self) -> str:
        """Alias for content_digest."""
        return self.content_digest

    _text_id = Column(Integer, db.ForeignKey(Blob.id), info=NOT_AUDITABLE)
    text_blob = relationship(Blob, cascade="all, delete", foreign_keys=[_text_id])

    _pdf_id = Column(Integer, db.ForeignKey(Blob.id), info=NOT_AUDITABLE)
    pdf_blob = relationship(Blob, cascade="all, delete", foreign_keys=[_pdf_id])

    _preview_id = Column(Integer, db.ForeignKey(Blob.id), info=NOT_AUDITABLE)
    preview_blob = relationship(Blob, cascade="all, delete", foreign_keys=[_preview_id])

    language = Column(
        Text, info={"searchable": True, "index_to": [("language", wf.ID(stored=True))]}
    )
    size = Column(Integer)
    page_num = Column(Integer, default=1)

    # FIXME: use Entity.meta instead
    #: Stores extra metadata as a JSON column
    extra_metadata_json = Column(UnicodeText, info={"auditable": False})

    sbe_type = "cmis:document"

    # antivirus status
    def ensure_antivirus_scheduled(self) -> None:
        if not self.antivirus_required:
            return

        # task_id = self.content_blob.meta.get("antivirus_task_id")
        task_id = self.id
        if task_id is not None:
            # if fail, the task will we tried again 20 times
            tasks.process_document.send(task_id)

    @property
    def antivirus_scanned(self) -> bool:
        """True if antivirus task was run, even if antivirus didn't return a
        result."""
        return self.content_blob and "antivirus" in self.content_blob.meta

    @property
    def antivirus_status(self) -> bool | None:
        """
        True: antivirus has scanned file: no virus
        False: antivirus has scanned file: virus detected
        None: antivirus task was run, but antivirus didn't return a result
        """
        return self.content_blob and self.content_blob.meta.get("antivirus")

    @property
    def antivirus_required(self) -> bool:
        """True if antivirus doesn't need to be run."""
        required = current_app.config["ANTIVIRUS_CHECK_REQUIRED"]
        return required and (
            not self.antivirus_scanned or self.antivirus_status is None
        )

    @property
    def antivirus_ok(self) -> bool:
        """True if user can safely access document content."""
        required = current_app.config["ANTIVIRUS_CHECK_REQUIRED"]
        if required:
            return self.antivirus_status is True

        return self.antivirus_status is not False

    # R/W properties
    @BaseContent.content.setter
    def content(self, value: bytes) -> None:
        BaseContent.content.fset(self, value)
        self.content_blob.meta["antivirus_task_id"] = str(uuid.uuid4())
        self.pdf_blob = None
        self.text_blob = None

    def set_content(self, content: bytes, content_type: str = "") -> None:
        super().set_content(content, content_type)
        async_conversion(self)

    @property
    def pdf(self):
        if self.pdf_blob:
            assert isinstance(self.pdf_blob.value, bytes)
        return self.pdf_blob and self.pdf_blob.value

    @pdf.setter
    def pdf(self, value: bytes) -> None:
        assert isinstance(value, bytes)
        self.pdf_blob = Blob()
        self.pdf_blob.value = value

    # `text` is an Unicode value.
    @property
    def text(self) -> str:
        return self.text_blob.value.decode("utf8") if self.text_blob is not None else ""

    @text.setter
    def text(self, value: str) -> None:
        assert isinstance(value, str)
        self.text_blob = Blob()
        self.text_blob.value = value.encode("utf8")

    @property
    def extra_metadata(self) -> dict[str, Any]:
        if not hasattr(self, "_extra_metadata"):
            if self._extra_metadata is not None:
                self._extra_metadata = json.loads(self.extra_metadata_json)
            else:
                self._extra_metadata = None
        return self._extra_metadata

    @extra_metadata.setter
    def extra_metadata(self, extra_metadata: dict[str, Any]) -> None:
        self._extra_metadata = extra_metadata
        self.extra_metadata_json = str(json.dumps(extra_metadata))

    # TODO: or use SQLAlchemy alias?
    @property
    def file_name(self) -> str:
        return self.title

    def __repr__(self):
        return "<Document id={!r} name={!r} path={!r} content_length={:d} at 0x{:x}>".format(
            self.id, self.title, self.path, self.content_length, id(self)
        )

    # locking management; used for checkin/checkout - this could be generalized to
    # any entity
    @property
    def lock(self) -> Lock | None:
        """
        :returns: either `None` if no lock or current lock is expired; either the
        current valid :class:`Lock` instance.
        """
        lock = self.meta.setdefault("abilian.sbe.documents", {}).get("lock")
        if lock:
            lock = Lock(**lock)
            if lock.expired:
                lock = None

        return lock

    @lock.setter
    def lock(self, user):
        """Allow to do `document.lock = user` to set a lock for user.

        If user is None, the lock is released.
        """
        if user is None:
            del self.lock
            return

        self.set_lock(user=user)

    @lock.deleter
    def lock(self):
        """Remove lock, if any.

        `del document.lock` can be safely done even if no lock is set.
        """
        meta = self.meta.setdefault("abilian.sbe.documents", {})
        if "lock" in meta:
            del meta["lock"]
            self.meta.changed()

    def set_lock(self, user=None):
        if user is None:
            user = current_user

        lock = self.lock
        if lock and not lock.is_owner(user=user):
            raise RuntimeError("This document is already locked by another user")

        meta = self.meta.setdefault("abilian.sbe.documents", {})
        lock = Lock.new()
        meta["lock"] = lock.as_dict()
        self.meta.changed()


def icon_for(content_type: str) -> str:
    for extension, mime_type in mimetypes.types_map.items():
        if mime_type == content_type:
            extension = extension[1:]
            icon = f"{extension}.png"
            if icon_exists(icon):
                return icon_url(icon)

    return icon_url("bin.png")


# Async conversion
_async_data = threading.local()


def _get_documents_queue() -> list[tuple[Document, str]]:
    if not hasattr(_async_data, "documents"):
        _async_data.documents = []
    return _async_data.documents


def async_conversion(document: Document) -> None:
    _get_documents_queue().append(
        # (document, document.content_blob.meta.get("antivirus_task_id"))
        (document, document.id)
    )


def _trigger_conversion_tasks(session: Session) -> None:
    if (
        # this commit is not from the application session
        session is not db.session()
        # inside a sub-transaction: not yet written in DB
        or session.transaction.nested
    ):
        return

    document_queue = _get_documents_queue()
    while document_queue:
        doc, task_id = document_queue.pop()
        if doc.id and isinstance(doc.id, int):
            logger.debug(
                "_trigger_conversion_tasks() doc.id={doc_id}",
                doc_id=repr(doc.id),
            )
            tasks.process_document.send(doc.id)


def setup_listener() -> None:
    mark_attr = "__abilian_sa_listening"
    if getattr(_trigger_conversion_tasks, mark_attr, False):
        return

    listen(Session, "after_commit", _trigger_conversion_tasks)
    setattr(_trigger_conversion_tasks, mark_attr, True)



""""""

from __future__ import annotations

from .base import (
    AUDITABLE,
    AUDITABLE_HIDDEN,
    EDITABLE,
    EXPORTABLE,
    NOT_AUDITABLE,
    NOT_EDITABLE,
    NOT_EXPORTABLE,
    NOT_SEARCHABLE,
    SEARCHABLE,
    SYSTEM,
    IdMixin,
    Indexable,
    Model,
    TimestampedMixin,
)
from .base_mixin import BaseMixin
from .owned import OwnedMixin
